:- module board.

:- interface.

:- import_module io, int, assoc_list, list, bt_array.
:- import_module ink, stream.

:- type board
    --->    board(
		width :: int,
		height :: int,
		things :: things,
		inkmaps :: assoc_list({int,int}, inkmap),
		grid :: grid
	    ).

:- type dir ---> n ; s ; e ; w.

:- type grid == bt_array(tile).

:- type things == assoc_list({int,int}, list(thing)).

:- type thing
    --->    package(int, int, int, int)	    % package(ID, X, Y, Weight)
    ;	    possible.

:- type tile
    --->    plain
    ;	    water
    ;	    wall.

% tile(X, Y) = Tile
:- func tile(int, int, board) = tile.

% passable(X, Y, Board)
:- pred passable(int, int, board).
:- mode passable(in, in, in) is semidet.

% passable_adjacent(X, Y, Board) = Adjacent
:- func passable_adjacent(int, int, board) = assoc_list(dir, {int, int}).

% read_board(In, Board)
:- pred read_board(In, board, io, io) <= putback(In).
:- mode read_board(in, out, di, uo) is det.

% write_board(Board)
:- pred write_board(board, io, io).
:- mode write_board(in, di, uo) is det.

% write_inked_board(Board)
:- pred write_inked_board(board, io, io).
:- mode write_inked_board(in, di, uo) is det.

%--------------------------------------------------------------------%

:- implementation.

:- import_module string, char, require, std_util.
:- import_module util.

tile(X, Y, Board) = lookup(Board ^ grid, Y*(Board ^ width + 2)+X).

passable(X, Y, Board) :- tile(X, Y, Board) = plain.
    
passable_adjacent(X, Y, Board) =
    ( if passable(X+1, Y, Board) then [e-{X+1, Y}] else [] ) ++
    ( if passable(X-1, Y, Board) then [w-{X-1, Y}] else [] ) ++
    ( if passable(X, Y+1, Board) then [n-{X, Y+1}] else [] ) ++
    ( if passable(X, Y-1, Board) then [s-{X, Y-1}] else [] ).

%--------------------------------------------------------------------%

read_board(In, Board) -->
    read_int(In, "board width", Width),
    read_int(In, "board height", Height),
    ignore_whitespace(In, _),
    read_board(In, Width, Height, Board).

%--------------------------------------------------------------------%
	
% read_board(In, Width, Height, Board)
:- pred read_board(In, int, int, board, io, io) <= putback(In).
:- mode read_board(in, in, in, out, di, uo) is det.

read_board(In, Width, Height, board(Width, Height, Things, [], Grid)) -->
    { Grid0 = init(0, (Height+2)*(Width+2), wall) },
    read_rows(In, Width, Height, 1, Things, Grid0, Grid).

% read_rows(In, Width, Height, Y, Things, Grid0, Grid)
:- pred read_rows(In, int, int, int, things, grid, grid, io, io) <= putback(In).
:- mode read_rows(in, in, in, in, out, in, out, di, uo) is det.

read_rows(In, Width, Height, Y, Things0++Things1, Grid0, Grid) -->
    read_line(In, Res0),
    { Res0 = ok(Line0), Line = from_char_list(Line0)
    ; Res0 = error(E), error(E)
    ; Res0 = eof, error("eof")
    },
    read_row(Width, 1, Y, Line, Things0, Grid0, Grid1),
    ( if { Y < Height } then
	read_rows(In, Width, Height, Y+1, Things1, Grid1, Grid)
    else
	{ Grid = Grid1, Things1 = [] }
    ).

% read_row(Width, X, Y, Line, Things, Grid0, Grid)
:- pred read_row(int, int, int, string, things, grid, grid, io, io).
:- mode read_row(in, in, in, in, out, in, out, di, uo) is det.

read_row(Width, X, Y, Line, Things0++Things1, Grid0, Grid) -->
    {
	( if index(Line, X-1, Char0) then
	    Char = Char0
	else
	    error("indexing '"++Line++"' at "++int_to_string(X-1))
	),
	(    if Char = ('#') then Tile = wall, Things0 = []
	else if Char = ('~') then Tile = water, Things0 = []
	else if Char = ('@') then Tile = plain, Things0 = [{X,Y}-[possible]]
	else			  Tile = plain, Things0 = []
	),
	set(Grid0, Y*(Width+2)+X, Tile, Grid1)
    },
    ( if { X < Width } then
	read_row(Width, X+1, Y, Line, Things1, Grid1, Grid)
    else
	{ Grid = Grid1, Things1 = [] }
    ).

%--------------------------------------------------------------------%

write_board(Board) -->
    write_rows(Board ^ width, Board ^ height, 0, Board ^ grid).

write_inked_board(Board) -->
    ( if { Board ^ inkmaps = [_-Inkmap|_] } then
	write_inked_rows(Inkmap, Board ^ width, Board ^ height, 0, Board ^ grid)
    else
	write_rows(Board ^ width, Board ^ height, 0, Board ^ grid)
    ).

:- pred write_rows(int, int, int, grid, io, io).
:- mode write_rows(in, in, in, in, di, uo) is det.

write_rows(Width, Height, Y, Grid) -->
    write_row(Width, Y, 0, Grid), nl,
    ( if { Y < Height+1 } then
	write_rows(Width, Height, Y+1, Grid)
    ).

:- pred write_inked_rows(inkmap, int, int, int, grid, io, io).
:- mode write_inked_rows(in, in, in, in, in, di, uo) is det.

write_inked_rows(Inkmap, Width, Height, Y, Grid) -->
    write_inked_row(Inkmap, Width, Y, 0, Grid), nl,
    ( if { Y < Height+1 } then
	write_inked_rows(Inkmap, Width, Height, Y+1, Grid)
    ).

:- pred write_row(int, int, int, grid, io, io).
:- mode write_row(in, in, in, in, di, uo) is det.

write_row(Width, Y, X, Grid) -->
    { lookup(Grid, Y*(Width+2)+X, Tile) },
    write_tile(Tile),
    ( if { X < Width+1 } then
	write_row(Width, Y, X+1, Grid)
    ).

:- pred write_inked_row(inkmap, int, int, int, grid, io, io).
:- mode write_inked_row(in, in, in, in, in, di, uo) is det.

write_inked_row(Inkmap, Width, Y, X, Grid) -->
    ( if { distance(Inkmap, {X,Y}) = D, int_to_digit(D, Char) } then
	write_char(Char)
    else
       { lookup(Grid, Y*(Width+2)+X, Tile) },
	write_tile(Tile)
    ),
    ( if { X < Width+1 } then
	write_inked_row(Inkmap, Width, Y, X+1, Grid)
    ).

:- pred write_tile(tile, io, io).
:- mode write_tile(in, di, uo) is det.

write_tile(plain) --> write_string(".").
write_tile(wall) --> write_string("#").
write_tile(water) --> write_string("~").
/*
write_inked_square(square(plain, Facts)) -->
    ( if { Facts = [_-Distance|_], int_to_digit(Distance, Char) } then
	write_char(Char)
    else
	write_string(".")
    ).
write_inked_square(square(wall, _)) --> write_string("#").
write_inked_square(square(water, _)) --> write_string("~").
*/
%--------------------------------------------------------------------%

:- end_module board.

