:- module ink.

:- interface.

:- import_module int, std_util, list.
:- import_module board.

:- type inkmap.
:- type distance == int.

% closest({X,Y}, Targets, Target, Board)
:- pred closest({int,int}, list({int,int}), maybe({int,int}), board).
:- mode closest(in, in, out, in) is det.

% distance({DX,DY}, {X,Y}, Distance, Board0, Board)
:- pred distance({int,int}, {int,int}, distance, board, board).
:- mode distance(in, in, out, in, out) is semidet.

% distance(Inkmap, {X,Y}) = Distance
:- func distance(inkmap, {int, int}) = distance is semidet.

% remove_all_but(Target, Board0, Board)
:- pred remove_all_but(maybe({int,int}), board, board).
:- mode remove_all_but(in, in, out) is det.

%--------------------------------------------------------------------%

:- implementation.

:- import_module queue, string, require, assoc_list, bt_array.

:- type inkmap
    --->    inkmap(
		width :: int,
		inks :: inks,
		grid :: bt_array(distance)
	    ).

:- type ink
    --->    ink(int, int, int).	% ink(X, Y, Distance)

:- type inks == queue(ink).

distance({DX,DY}, {X,Y}, D, Board0, Board) :-
    ( if search(Board0 ^ inkmaps, {DX,DY}, Inkmap) then
	D = distance(Inkmap, {X,Y}),
	Board = Board0
    else
	add_inkmap({DX,DY}, never, _, Board0, Board1),
	distance({DX,DY}, {X,Y}, D, Board1, Board)
    ).

:- pred never({int,int}).
:- mode never(in) is semidet.

never(_) :- semidet_fail.

distance(Inkmap, {X,Y}) = D :-
    lookup(Inkmap ^ grid, Y*(Inkmap ^ width+2)+X, D),
    D >= 0.

:- pred considering(inkmap, {int, int}).
:- mode considering(in, in) is semidet.

considering(Inkmap, {X, Y}) :- lookup(Inkmap ^ grid, Y*(Inkmap ^ width+2)+X, -2).

:- pred set_considering({int, int}, inkmap, inkmap).
:- mode set_considering(in, in, out) is det.

set_considering({X,Y}, Inkmap0, Inkmap) :-
    set(Inkmap0 ^ grid, Y*(Inkmap0^width+2)+X, -2, Grid),
    Inkmap = Inkmap0 ^ grid := Grid.

:- pred set_distance(int, int, distance, inkmap, inkmap).
:- mode set_distance(in, in, in, in, out) is det.

set_distance(X, Y, D, Inkmap0, Inkmap) :-
    ( if distance(Inkmap0, {X,Y}) = _ then
	error(int_to_string(X)++","++int_to_string(Y)++
	    " already has a distance "++int_to_string(D))
    else
	Inkmap1 = Inkmap0
    ),
    set(Inkmap1 ^ grid, Y*(Inkmap1 ^ width+2)+X, D, Grid),
    Inkmap = Inkmap1 ^ grid := Grid.

%--------------------------------------------------------------------%

remove_all_but(MaybeTarget, Board0, Board) :-
    ( if MaybeTarget = yes(Target), search(Board0 ^ inkmaps, Target, Inkmap) then
	Board = Board0 ^ inkmaps := [Target-Inkmap]
    else
	Board = Board0 ^ inkmaps := []
    ).

%--------------------------------------------------------------------%

closest({X,Y}, Targets, Target, Board0) :-
    add_inkmap({X,Y}, match(Targets), Target, Board0, _).

:- pred match(list({int,int}), {int,int}).
:- mode match(in, in) is semidet.

match(Targets, Target) :- member(Target, Targets).

%--------------------------------------------------------------------%

% add_inkmap({X,Y}, Pred, Target, Board0, Board)
:- pred add_inkmap({int,int}, pred({int,int}), maybe({int,int}), board, board).
:- mode add_inkmap(in, pred(in) is semidet, out, in, out) is det.

add_inkmap({X,Y}, Pred, Target, Board0, Board) :-
    W = Board0 ^ width,
    H = Board0 ^ height,
    Inkmap0 = inkmap(W, list_to_queue([ink(X,Y,0)]), init(0, (W+2)*(H+2), -1)),
    set_considering({X,Y}, Inkmap0, Inkmap1),
    ink_board(Board0, Pred, Target, Inkmap1, Inkmap),
    Board = Board0 ^ inkmaps := [{X,Y}-Inkmap|Board0 ^ inkmaps].

% ink_board(Board, Pred, Target, Inkmap0, Inkmap)
:- pred ink_board(board, pred({int,int}), maybe({int,int}), inkmap, inkmap).
:- mode ink_board(in, pred(in) is semidet, out, in, out) is det.

ink_board(Board, Pred, Target, Inkmap0, Inkmap) :-
    ( if get(Inkmap0 ^ inks, ink(X, Y, Distance), Q) then
	set_distance(X, Y, Distance, Inkmap0 ^ inks := Q, Inkmap1),
	( if Pred({X,Y}) then
	    Target = yes({X,Y}),
	    Inkmap = Inkmap1
	else
	    consider(X+1, Y, Distance+1, Board, Inkmap1, Inkmap2),
	    consider(X-1, Y, Distance+1, Board, Inkmap2, Inkmap3),
	    consider(X, Y+1, Distance+1, Board, Inkmap3, Inkmap4),
	    consider(X, Y-1, Distance+1, Board, Inkmap4, Inkmap5),
	    ink_board(Board, Pred, Target, Inkmap5, Inkmap)
	)
    else
	Inkmap = Inkmap0,
	Target = no
    ).

% consider(X, Y, Distance, Board, Inkmap0, Inkmap)
:- pred consider(int, int, int, board, inkmap, inkmap).
:- mode consider(in, in, in, in, in, out) is det.

consider(X, Y, Distance, Board, Inkmap0, Inkmap) :-
    ( if
	passable(X, Y, Board),
	not distance(Inkmap0, {X,Y}) = _,
	not considering(Inkmap0, {X,Y})
    then
	put(Inkmap0 ^ inks, ink(X, Y, Distance), Q),
	set_considering({X,Y}, Inkmap0 ^ inks := Q, Inkmap)
    else
	Inkmap = Inkmap0
    ).

%--------------------------------------------------------------------%

:- end_module ink.

