:- module posix__stream.		% Copyright (C) 2002 YesLogic Pty. Ltd.

:- interface.

:- import_module stream.

:- type input_fd.
:- type output_fd.
:- type duplex_fd.

:- instance input(input_fd).
:- instance output(output_fd).
:- instance input(duplex_fd).
:- instance output(duplex_fd).
:- instance duplex(duplex_fd).

% create_input(Fd, Stream)
:- pred create_input(fd, input_fd, io, io).
:- mode create_input(in, out, di, uo) is det.

% create_output(Fd, Stream)
:- pred create_output(fd, output_fd, io, io).
:- mode create_output(in, out, di, uo) is det.

% create_duplex(Fd, Stream)
:- pred create_duplex(fd, duplex_fd, io, io).
:- mode create_duplex(in, out, di, uo) is det.

%--------------------------------------------------------------------%

:- implementation.

:- import_module posix__buffer.
:- import_module posix__read.
:- import_module posix__write.
:- import_module mvar, require.

:- type stream
    --->    stream(
		offset :: mvar(int),
		available :: mvar(int),
		size :: int,
		buffer :: pointer
	    ).

:- type input_fd
    --->    input_fd(stream, fd).

:- type output_fd
    --->    output_fd(stream, fd).

:- type duplex_fd
    --->    duplex_fd(
		input :: stream,
		output :: stream,
		fd :: fd
	    ).

%--------------------------------------------------------------------%

:- instance input(input_fd) where
    [
	input_name(input_fd(_, fd(Fd)), 
	    "<<fd "++int_to_string(Fd)++">>") --> [],
	
	(read_char(input_fd(Stream, Fd), Res) -->
	    { Stream = stream(Offset, Avail, Size, Buffer) },
	    take(Offset, I),
	    take(Avail, A),
	    ( if { I < A } then
		index(I, Buffer, Char),
		put(Offset, I+1),
		put(Avail, A),
		{ Res = ok(Char) }
	    else
		read(Fd, Buffer, Size, Res0),
		(
		    { Res0 = ok(Count) },
		    put(Avail, Count),
		    put(Offset, 0),
		    read_char(input_fd(Stream, Fd), Res)
		;
		    { Res0 = error(_Error), Res = error("foo") }, %FIXME
		    put(Avail, 0),
		    put(Offset, 0)
		)
	    )
	)
    ].

%--------------------------------------------------------------------%

:- instance output(output_fd) where
    [
	output_name(output_fd(_, fd(Fd)), 
	    "<<fd "++int_to_string(Fd)++">>") --> [],
	
	(write_char(output_fd(Stream, Fd), Char) -->
	    { Stream = stream(_Offset, Avail, Size, Buffer) },
	    take(Avail, A),
	    set(A, Buffer, Char),
	    put(Avail, A+1),
	    ( if { A > Size - 1 ; Char = '\n' } then
		flush_output(output_fd(Stream, Fd))
	    )
	)
    ].

:- pred flush_output(output_fd, io, io).
:- mode flush_output(in, di, uo) is det.

flush_output(output_fd(Stream, Fd)) -->
    { Stream = stream(Offset, Avail, _Size, Buffer) },
    take(Offset, I),
    take(Avail, A),
    offset(I, Buffer, NewBuffer),
    write(Fd, NewBuffer, A - I, Res0),
    (
	{ Res0 = ok(Count) },
	( if { Count + I >= A } then
	    put(Offset, 0),
	    put(Avail, 0)
	else
	    put(Offset, I+Count),
	    put(Avail, A),
	    flush_output(output_fd(Stream, Fd))
	)
    ;
	{ Res0 = error(_Error), error("foo") }, %FIXME
	put(Offset, I),
	put(Avail, A)
    ).

%--------------------------------------------------------------------%

create_input(Fd, input_fd(Stream, Fd)) --> create_stream(Stream).

create_output(Fd, output_fd(Stream, Fd)) --> create_stream(Stream).

create_duplex(Fd, duplex_fd(S1, S2, Fd)) -->
    create_stream(S1), create_stream(S2).

:- pred create_stream(stream, io, io).
:- mode create_stream(out, di, uo) is det.

create_stream(stream(Offset, Avail, 8192, Buffer)) -->
    init(Offset),
    init(Avail),
    put(Offset, 0),
    put(Avail, 0),
    alloc(8192, Buffer).

%--------------------------------------------------------------------%

:- end_module posix__stream.

