:- module stream.

:- interface.

:- import_module char, io, string, list, std_util.

:- type stream__result(T)
	--->	ok(T)
	;	eof
	;	error(string).

:- type stream__result
	--->	ok
	;	eof
	;	error(string).

:- type stream__res
	--->	ok
	;	error(string).

%--------------------------------------------------------------------%

:- typeclass input(S) where
    [
	pred input_name(S, string, io, io),
	mode input_name(in, out, di, uo) is det,

	pred read_char(S, stream__result(char), io, io),
	mode read_char(in, out, di, uo) is det
    ].

:- typeclass output(S) where
    [
	pred output_name(S, string, io, io),
	mode output_name(in, out, di, uo) is det,

	pred write_char(S, char, io, io),
	mode write_char(in, in, di, uo) is det
    ].

:- typeclass duplex(S) <= (input(S), output(S)) where [].

%--------------------------------------------------------------------%

:- typeclass putback(S) <= input(S) where
    [
	pred putback_char(S, char, io, io),
	mode putback_char(in, in, di, uo) is det
    ].

:- typeclass unbounded_putback(S) <= putback(S) where [].

:- typeclass line(S) <= input(S) where
    [
	pred line_number(S, int, io, io),
	mode line_number(in, out, di, uo) is det,

	pred set_line_number(S, int, io, io),
	mode set_line_number(in, in, di, uo) is det
    ].

%--------------------------------------------------------------------%

:- pred read_line(S, stream__result(list(char)), io, io) <= input(S).
:- mode read_line(in, out, di, uo) is det.

%--------------------------------------------------------------------%

% Reads a whitespace delimited word from the input stream
:- pred read_word(S, stream__result(list(char)), io, io) <= putback(S).
:- mode read_word(in, out, di, uo) is det.

% Discards all the whitespace from the input stream
:- pred ignore_whitespace(S, stream__result, io, io) <= putback(S).
:- mode ignore_whitespace(in, out, di, uo) is det.

%--------------------------------------------------------------------%

% Write the string to the output stream
:- pred write_string(S, string, io, io) <= output(S).
:- mode write_string(in, in, di, uo) is det.

%--------------------------------------------------------------------%

% Echo stream In to stream Out.
:- pred cat(In, Out, stream__res, io, io) <= (input(In), output(Out)).
:- mode cat(in, in, out, di, uo) is det.

%--------------------------------------------------------------------%
% An input stream with infinite putback

:- type putback(S).
:- instance input(putback(S)) <= input(S).
:- instance putback(putback(S)) <= input(S).

:- pred putback_stream(S, putback(S), io, io) <= input(S).
:- mode putback_stream(in, out, di, uo) is det.

:- func putback_base_stream(putback(S)) = S <= input(S).

%-----------------------------------------------------------------------------%
% An input stream that numbers lines (starting from one)

:- type linenumber(S).
:- instance input(linenumber(S)) <= input(S).
:- instance putback(linenumber(S)) <= putback(S).
:- instance line(linenumber(S)) <= input(S).

:- pred linenumber_stream(S, linenumber(S), io, io) <= input(S).
:- mode linenumber_stream(in, out, di, uo) is det.

:- func linenumber_base_stream(linenumber(S)) = S <= input(S).

%-----------------------------------------------------------------------------%

:- instance input(input_stream).
:- instance putback(input_stream).

% with_stdin(Pred, Res)
:- pred with_stdin(pred(input_stream, T, io, io), T, io, io).
:- mode with_stdin(pred(in, out, di, uo) is det, out, di, uo) is det.

% with_file(FileName, Pred, Res)
:- pred with_file(string, pred(input_stream, maybe_error(T), io, io),
					maybe_error(T), io, io).
:- mode with_file(in, pred(in, out, di, uo) is det, out, di, uo) is det.

%-----------------------------------------------------------------------------%

:- instance output(output_stream).

%-----------------------------------------------------------------------------%

:- implementation.

:- import_module int.
:- import_module mvar.

:- instance input(input_stream) where
    [
	pred(input_name/4) is input_stream_name,

	(read_char(S, Res) -->
	    read_char(S, Res0),
	    {
		Res0 = ok(Char), Res = ok(Char)
	    ;
		Res0 = eof, Res = eof
	    ;
		Res0 = error(Error), Res = error(error_message(Error))
	    }
	)
    ].

:- instance putback(input_stream) where
    [
	pred(putback_char/4) is io__putback_char
    ].

with_stdin(Pred, Res) -->
    stdin_stream(Stdin),
    Pred(Stdin, Res).

with_file(FileName, Pred, Res) -->
    open_input(FileName, Res0),
    (
        { Res0 = ok(Stream) },
	Pred(Stream, Res),
	close_input(Stream)
    ;
        { Res0 = error(Error), Res = error(error_message(Error)) }
    ).

%--------------------------------------------------------------------%

:- instance output(output_stream) where
    [
	pred(output_name/4) is output_stream_name,
	pred(write_char/4) is io__write_char
    ].

%--------------------------------------------------------------------%

:- type putback(S)
    --->    pb(S, mvar(list(char))).

:- instance input(putback(S)) <= input(S) where
    [
	input_name(pb(S, _), Name) --> input_name(S, Name),
	pred(read_char/4) is putback_read_char
    ].

:- instance putback(putback(S)) <= input(S) where
    [
	pred(stream__putback_char/4) is putback_putback_char
    ].

putback_stream(Stream, pb(Stream, MPutbackChars)) -->
	mvar__init(MPutbackChars),
	mvar__put(MPutbackChars, []).

putback_base_stream(pb(Stream, _)) = Stream.

:- pred putback_read_char(putback(S)::in, stream__result(char)::out,
		io__state::di, io__state::uo) is det <= stream__input(S).

putback_read_char(pb(Stream, MPutbackChars), Result) -->
	mvar__take(MPutbackChars, PutbackChars),
	(
		{ PutbackChars = [] },
		{ NewPutbackChars = PutbackChars },
		stream__read_char(Stream, Result)
	;
		{ PutbackChars = [Char | NewPutbackChars] },
		{ Result = ok(Char) }
	),
	mvar__put(MPutbackChars, NewPutbackChars).

:- pred putback_putback_char(putback(S)::in, char::in,
		io__state::di, io__state::uo) is det <= stream__input(S).

putback_putback_char(pb(_Stream, MPutbackChars), Char) -->
	mvar__take(MPutbackChars, PutbackChars),
	mvar__put(MPutbackChars, [Char | PutbackChars]).

%--------------------------------------------------------------------%

:- type linenumber(S)
	--->	line(
			S,		% stream
			mvar(int)	% line number
		).

:- instance stream__input(linenumber(S)) <= stream__input(S) where [
	(input_name(line(S, _), Name) --> input_name(S, Name)),
	pred(stream__read_char/4) is linenumber_read_char
].

:- instance stream__putback(linenumber(S)) <= stream__putback(S) where [
	pred(stream__putback_char/4) is linenumber_putback_char
].

:- instance stream__line(linenumber(S)) <= stream__input(S) where [
	pred(stream__line_number/4) is linenumber,
	pred(stream__set_line_number/4) is set_linenumber
].

linenumber_stream(S, line(S, MLine)) -->
	mvar__init(MLine),
	mvar__put(MLine, 0).

linenumber_base_stream(line(Stream, _)) = Stream.

:- pred linenumber_read_char(linenumber(S)::in, stream__result(char)::out,
		io__state::di, io__state::uo) is det <= stream__input(S).

linenumber_read_char(line(Stream, MLine), Result) -->
	stream__read_char(Stream, Result),
	( { Result = ok('\n') } ->
		mvar__take(MLine, Line),
		mvar__put(MLine, Line + 1)
	;
		[]
	).

:- pred linenumber_putback_char(linenumber(S)::in, char::in,
		io__state::di, io__state::uo) is det <= stream__putback(S).

linenumber_putback_char(line(Stream, MLine), Char) -->
	stream__putback_char(Stream, Char),
	( { Char = '\n' } ->
		mvar__take(MLine, Line),
		mvar__put(MLine, Line - 1)
	;
		[]
	).

:- pred linenumber(linenumber(S)::in, int::out,
		io__state::di, io__state::uo) is det.

linenumber(line(_, MLine), Line) -->
	mvar__take(MLine, Line),
	mvar__put(MLine, Line).

:- pred set_linenumber(linenumber(S)::in, int::in,
		io__state::di, io__state::uo) is det.

set_linenumber(line(_, MLine), Line) -->
	mvar__take(MLine, _OldLine),
	mvar__put(MLine, Line).

%--------------------------------------------------------------------%

read_line(Stream, Res) -->
    read_char(Stream, Res0),
    (
	{ Res0 = ok(Char) },
	( if { Char = '\n' } then
	    { Res = ok([Char]) }
	else
	    read_line(Stream, Res1),
	    (
		{ Res1 = ok(Chars), Res = ok([Char|Chars]) }
	    ;
		{ Res1 = eof, Res = ok([Char]) }
	    ;
		{ Res1 = error(Error), Res = error(Error) }
	    )
	)	
    ;
	{ Res0 = eof, Res = eof }
    ;
	{ Res0 = error(Error), Res = error(Error) }
    ).

%--------------------------------------------------------------------%

read_word(Stream, Res) -->
    ignore_whitespace(Stream, Res0),
    (
	{ Res0 = ok },
	read_word_2(Stream, Res)
    ;
	{ Res0 = eof, Res = eof }
    ;
	{ Res0 = error(Error), Res = error(Error) }
    ).

:- pred read_word_2(S, stream__result(list(char)), io, io) <= putback(S).
:- mode read_word_2(in, out, di, uo) is det.

read_word_2(Stream, Res) -->
    read_char(Stream, Res0),
    (
	{ Res0 = ok(Char) },
	( if { is_whitespace(Char) } then
	    putback_char(Stream, Char),
	    { Res = ok([]) }
	else
	    read_word_2(Stream, Res1),
	    (
		{ Res1 = ok(Chars), Res = ok([Char|Chars]) }
	    ;
		{ Res1 = eof, Res = ok([Char]) }
	    ;
		{ Res1 = error(Error), Res = error(Error) }
	    )
	)	
    ;
	{ Res0 = eof, Res = eof }
    ;
	{ Res0 = error(Error), Res = error(Error) }
    ).

ignore_whitespace(Stream, Res) -->
    read_char(Stream, Res0),
    (
	{ Res0 = ok(Char) },
	( if { is_whitespace(Char) } then
	    ignore_whitespace(Stream, Res)
	else
	    putback_char(Stream, Char),
	    { Res = ok }
	)	
    ;
	{ Res0 = eof, Res = eof }
    ;
	{ Res0 = error(Error), Res = error(Error) }
    ).

%--------------------------------------------------------------------%

write_string(Stream, String) --> foldl(write_char(Stream), String).
	
%--------------------------------------------------------------------%

cat(In, Out, Res) -->
    read_char(In, Res0),
    (
	{ Res0 = ok(Char) },
	write_char(Out, Char),
	cat(In, Out, Res)
    ;
	{ Res0 = eof, Res = ok }
    ;
	{ Res0 = error(Error), Res = error(Error) }
    ).

%--------------------------------------------------------------------%

:- end_module stream.

