:- module world.

:- interface.

:- import_module int, io, map, set, random, std_util.
:- import_module board, stream.

:- type world
    --->    world(
		id :: int,
		capacity :: int,
		money :: int,
		target :: maybe({int,int}),
		robots :: robots,
		packages :: map(int, pinfo),
		supply :: supply,
		board :: board
	    ).

:- type robots == map(int, robot).

:- type robot
    --->    robot(
		x :: int,
		y :: int,
		carrying :: set(int)
	    ).

:- type pinfo
    --->    pinfo(int, int, int).   % pinfo(X, Y, W)

% read_world(In, World)
:- pred read_world(In, world, io, io) <= putback(In).
:- mode read_world(in, out, di, uo) is det.

% go(In, Out, World0, World)
:- pred go(In, Out, world, world, io, io) <= (putback(In), output(Out)).
:- mode go(in, in, in, out, di, uo) is det.

%--------------------------------------------------------------------%

:- implementation.

:- import_module list, assoc_list, string, require, char.
:- import_module ink, util.

read_world(In, World, !IO) :-
    stderr_stream(Stderr, !IO),
%    io__write_string(Stderr, "about to read board\n", !IO),
    read_board(In, Board, !IO),
%    io__write_string(Stderr, "finished reading board\n", !IO),
    read_int(In, "id", ID, !IO),
    read_int(In, "capacity", Capacity, !IO),
    read_int(In, "money", Money, !IO),
%    io__write_string(Stderr, "robot id="++int_to_string(ID)++"\n", !IO),
%    io__write_string(Stderr, "reading state\n", !IO),
    ignore_whitespace(In, _, !IO),
    time(T, !IO),
    init(T, Supply),
    World0 = world(ID, Capacity, Money, no, init, init, Supply, Board),
    read_state(In, World0, World1, !IO),
    next_target(World1, World).
%    io__write_string(Stderr, "done\n", !IO).

%--------------------------------------------------------------------%

:- pred read_state(In, world, world, io, io) <= putback(In).
:- mode read_state(in, in, out, di, uo) is det.

read_state(In, World0, World, !IO) :-
    read_line(In, Res0, !IO),
    ( if Res0 = ok(Line0) then
	Line = from_char_list(Line0),
	Updates0 = words(unify('#'), Line),
	Updates1 = map(words(is_whitespace), Updates0),
	Updates2 = filter_map(get_update, Updates1),
	foldl(apply_updates, Updates2, World0 ^ robots, Robots),
	World = World0 ^ robots := Robots
    else
	error("could not read state line")
    ).

:- type update
    --->    moved(dir)
    ;	    x(int)
    ;	    y(int)
    ;	    picked(int)
    ;	    dropped(int).

:- pred apply_updates(pair(int,list(update)), robots, robots).
:- mode apply_updates(in, in, out) is det.

apply_updates(Id-Updates, Robots0, Robots) :-
    foldl(apply_update(Id), Updates, Robots0, Robots).

:- pred apply_update(int, update, robots, robots).
:- mode apply_update(in, in, in, out) is det.

apply_update(RID, Update, Robots0, Robots) :-
    ( if search(Robots0, RID, R0) then
	R1 = R0
    else
	R1 = robot(1, 1, init)
    ),
    (
	Update = moved(n), R = R1 ^ y := R1 ^ y + 1,
	set(Robots0, RID, R, Robots)
    ;
	Update = moved(s), R = R1 ^ y := R1 ^ y - 1,
	set(Robots0, RID, R, Robots)
    ;
	Update = moved(e), R = R1 ^ x := R1 ^ x + 1,
	set(Robots0, RID, R, Robots)
    ;
	Update = moved(w), R = R1 ^ x := R1 ^ x - 1,
	set(Robots0, RID, R, Robots)
    ;
	Update = x(X), R = R1 ^ x := X,
	set(Robots0, RID, R, Robots)
    ;
	Update = y(Y), R = R1 ^ y := Y,
	set(Robots0, RID, R, Robots)
    ;
	Update = picked(ID),
	no_one_has_this_package(ID, Robots0, Robots1),
	insert(R1 ^ carrying, ID, Carrying),
	R = R1 ^ carrying := Carrying,
	set(Robots1, RID, R, Robots)
    ;
	Update = dropped(ID),
	no_one_has_this_package(ID, Robots0, Robots)
    ).

:- func get_update(list(string)) = pair(int, list(update)) is semidet.

get_update([ID0|Updates]) = ID-get_updates(Updates) :- to_int(ID0, ID).

:- func get_updates(list(string)) = list(update) is semidet.

get_updates([]) = [].
get_updates(["N"|Us]) = [moved(n)|get_updates(Us)].
get_updates(["S"|Us]) = [moved(s)|get_updates(Us)].
get_updates(["E"|Us]) = [moved(e)|get_updates(Us)].
get_updates(["W"|Us]) = [moved(w)|get_updates(Us)].
get_updates(["X",X0|Us]) = [x(X)|get_updates(Us)] :- to_int(X0, X).
get_updates(["Y",Y0|Us]) = [y(Y)|get_updates(Us)] :- to_int(Y0, Y).
get_updates(["P",ID0|Us]) = [picked(ID)|get_updates(Us)] :- to_int(ID0, ID).
get_updates(["D",ID0|Us]) = [dropped(ID)|get_updates(Us)] :- to_int(ID0, ID).

%--------------------------------------------------------------------%

:- func our_x(world) = int.
:- func our_y(world) = int.
:- func our_carrying(world) = set(int).

our_x(World) = Robot ^ x :- lookup(World ^ robots, World ^ id, Robot).

our_y(World) = Robot ^ y :- lookup(World ^ robots, World ^ id, Robot).

our_carrying(World) = Robot ^ carrying :-
    lookup(World ^ robots, World ^ id, Robot).

%--------------------------------------------------------------------%

go(In, Out, World0, World, !IO) :-    
    read_packages(In, World0, World1, !IO),
    interesting_packages(Packages, World1),

    ( if Packages \= [] then
	pick(Out, Packages, !IO),
	keep_going(In, Out, World1, World, !IO)
    else if should_drop(World1) = Droppings, Droppings \= [] then
	drop(Out, Droppings, !IO),
	keep_going(In, Out, World1, World, !IO)
    else 
	X = World1 ^ our_x,
	Y = World1 ^ our_y,
	( if World0 ^ target = yes({X, Y}) ; World0 ^ target = no then
	    next_target(World1, World2)
	else
	    World2 = World1
	),
	(
	    World2 ^ target = yes(Target),
	    move_to(In, Out, Target, World2, World, !IO)
	;
	    World2 ^ target = no,
	    noop(Out, !IO),
	    keep_going(In, Out, World2, World, !IO)
	)
    ).

:- pred move_to(In, Out, {int,int}, world, world, io, io) <= (putback(In), output(Out)).
:- mode move_to(in, in, in, in, out, di, uo) is det.

move_to(In, Out, Target, World0, World, !IO) :-
    stderr_stream(Stderr, !IO),
    Board0 = World0 ^ board,
    X = World0 ^ our_x,
    Y = World0 ^ our_y,
    Adj = passable_adjacent(X, Y, Board0),
    Risks = map(risk(World0), values(Adj)),
    ( if map_foldl(distance(Target), values(Adj), Distances, Board0, Board1) then
	DistAdj = from_corresponding_lists(Distances, keys(Adj)),
	RiskDistDir = from_corresponding_lists(Risks, DistAdj),
	permutation(RiskDistDir, RandRiskDistDir, World0 ^ supply, Supply0),
	SortedRiskDistDir = sort(RiskDistDir),
	World1 = World0 ^ board := Board1,
	( if SortedRiskDistDir = [LowRisk-_|_],
	    random(N, Supply0, Supply),
	    RandRisk0 = N mod 100,
	    ( if RandRisk0 < 60 then RandRisk = 0
	    else if RandRisk0 < 80 then RandRisk = 1
	    else if RandRisk0 < 95 then RandRisk = 2
	    else RandRisk = 3
	    ),
	    Risk = max(LowRisk, RandRisk),
	    DistDir = take_risk(Risk, RandRiskDistDir),
	    sort(DistDir) = [_-Dir|_]
	then
	    move(Out, Dir, !IO),
	    keep_going(In, Out, World1 ^ supply := Supply, World, !IO)
	else
%	    io__write_string(Stderr, "Boxed in!\n", !IO),
	    World = World1 ^ supply := Supply0
	)
    else
%	io__write_string(Stderr, "Cannot reach target!\n", !IO),
	World = World0
    ).

:- pred keep_going(In, Out, world, world, io, io) <= (putback(In), output(Out)).
:- mode keep_going(in, in, in, out, di, uo) is det.

keep_going(In, Out, World0, World, !IO) :-
    read_state(In, World0, World1, !IO),
%    dump_state(World1, !IO),
    go(In, Out, World1, World, !IO).

:- pred dump_state(world, io, io).
:- mode dump_state(in, di, uo) is det.

dump_state(World, !IO) :-
    stderr_stream(Stderr, !IO),
    foldl(dump_robot(World ^ robots), World ^ robots ^ keys, !IO),
    io__nl(Stderr, !IO).

:- pred dump_robot(robots, int, io, io).
:- mode dump_robot(in, in, di, uo) is det.

dump_robot(Robots, Id, !IO) :-
    lookup(Robots, Id, robot(X, Y, _)),
    stderr_stream(Stderr, !IO),
    io__write_string(Stderr,
	" #" ++ int_to_string(Id) ++ " " ++
	"("++int_to_string(X)++
	","++int_to_string(Y) ++ ")", !IO).

%--------------------------------------------------------------------%

:- func take_risk(int, assoc_list(int, T)) = list(T).

take_risk(_, []) = [].
take_risk(Risk, [R-T|Rs]) =
    ( if Risk >= R then
	[T|take_risk(Risk, Rs)]
    else
	take_risk(Risk, Rs)
    ).

:- func risk(world, {int,int}) = int.

risk(World, {X,Y}) = foldl(check_risk(X,Y), Robots, 0) :-
    Ids0 = World ^ robots ^ keys,
    delete_all(Ids0, World ^ id, Ids),
    Robots = map(lookup(World ^ robots), Ids).

:- func check_risk(int, int, robot, int) = int.

check_risk(X, Y, robot(RX, RY, _), Curr) = Risk :-
    ( if X = RX, Y = RY then		NewRisk = 3
    
    else if X - 1 = RX, Y = RY then	NewRisk = 2
    else if X + 1 = RX, Y = RY then	NewRisk = 2
    else if X = RX, Y + 1 = RY then	NewRisk = 2
    else if X = RX, Y - 1 = RY then	NewRisk = 2

    else if X + 2 = RX, Y = RY then	NewRisk = 1
    else if X - 2 = RX, Y = RY then	NewRisk = 1
    else if X - 1 = RX, Y - 1 = RY then	NewRisk = 1
    else if X + 1 = RX, Y - 1 = RY then	NewRisk = 1
    else if X - 1 = RX, Y + 1 = RY then	NewRisk = 1
    else if X + 1 = RX, Y + 1 = RY then	NewRisk = 1
    else if X = RX, Y + 2 = RY then	NewRisk = 1
    else if X = RX, Y - 2 = RY then	NewRisk = 1
	
    else				NewRisk = 0
    ),
    Risk = max(Curr, NewRisk).

% ..1..
% .121.
% 12321
% .121.
% ..1..

%--------------------------------------------------------------------%

:- pred next_target(world, world).
:- mode next_target(in, out) is det.

next_target(World0, World) :-
    Packages = filter_map(search(World0 ^ packages), to_sorted_list(World0 ^
	our_carrying)),
    ( if Packages = [_|_] then
	closest(
	    {World0 ^ our_x, World0 ^ our_y},
	    map(get_destination, Packages),
	    Target,
	    World0 ^ board
	),
	World1 = World0 ^ target := Target
    else if World0 ^ board ^ things = [_-_|_] then
	closest(
	    {World0 ^ our_x, World0 ^ our_y},
	    World0 ^ board ^ things ^ keys,
	    Target,
	    World0 ^ board
	),
	World1 = World0 ^ target := Target
    else
	World1 = World0 ^ target := no
    ),
    remove_all_but(World1 ^ target, World1 ^ board, Board),
    World = World1 ^ board := Board.

:- func get_destination(pinfo) = {int,int}.

get_destination(pinfo(X,Y,_)) = {X,Y}.

%--------------------------------------------------------------------%

:- pred interesting_packages(list(int), world).
:- mode interesting_packages(out, in) is det.

interesting_packages(Ps, World) :-
    X = World ^ our_x,
    Y = World ^ our_y,
    Capacity = current_capacity(World),
    Things0 = World ^ board ^ things,
    ( if search(Things0, {X,Y}, Things) then
	Ps = filter_map(interesting(Capacity), Things)
    else
	Ps = []
    ).

:- func interesting(int, thing) = int is semidet.

interesting(Capacity, package(ID, _X, _Y, W)) = ID :- W =< Capacity.

%--------------------------------------------------------------------%

:- pred read_packages(In, world, world, io, io) <= putback(In).
:- mode read_packages(in, in, out, di, uo) is det.

read_packages(In, World0, World, !IO) :-
    read_line(In, LineRes, !IO),
    ( if LineRes = ok(Line) then
	Ps0 = words(is_whitespace, from_char_list(Line)),
	filter_map(to_int, Ps0, Ps1),
	X = World0 ^ our_x,
	Y = World0 ^ our_y,
	Things0 = World0 ^ board ^ things,
	( if remove(Things0, {X,Y}, _, Things1) then
	    Things2 = Things1
	else
	    Things2 = Things0
	),
	stderr_stream(Stderr, !IO),
	( if Ps = get_packages(Ps1), Ps \= [] then
	    foldl(no_one_has_this_package, filter_map(get_thing_id, Ps), World0 ^ robots, Robots),
	    World1 = World0 ^ robots := Robots,
	    Things = [{X,Y}-Ps|Things2],
	    update_packages(Ps, World1, World2)
%	    io__write_string(Stderr, int_to_string(length(Ps))++" packages\t", !IO)
	else
%	    io__write_string(Stderr, "no packages\t", !IO),
	    World2 = World0,
	    Things = Things2
	),
	Board0 = World2 ^ board,
	Board = Board0 ^ things := Things,
	World = World2 ^ board := Board
    else
	error("no package line")
    ).

:- pred update_packages(list(thing), world, world).
:- mode update_packages(in, in, out) is det.

update_packages([]) --> [].
update_packages([possible|Ps]) --> update_packages(Ps).
update_packages([package(ID, X, Y, W)|Ps], World0, World) :-
    Packages0 = World0 ^ packages,
    set(Packages0, ID, pinfo(X, Y, W), Packages),
    World1 = World0 ^ packages := Packages,
    update_packages(Ps, World1, World).

:- func get_thing_id(thing) = int is semidet.

get_thing_id(package(ID,_,_,_)) = ID.

:- func get_packages(list(int)) = list(thing) is semidet.

get_packages([]) = [].
get_packages([ID,X,Y,W|Ps]) = [package(ID,X,Y,W)|get_packages(Ps)].

:- func current_capacity(world) = int.

current_capacity(World) = World ^ capacity - Weight :-
    Weights = map(get_package_weight(World ^ packages),
		to_sorted_list(World ^ our_carrying)),
    Weight = foldl(plus, Weights, 0).

:- func get_package_weight(map(int, pinfo), int) = int.

get_package_weight(Packages, ID) = Weight :-
    ( if search(Packages, ID, pinfo(_, _, W)) then
	Weight = W
    else
	error("unknown package")
    ).

:- pred no_one_has_this_package(int, robots, robots).
:- mode no_one_has_this_package(in, in, out) is det.

no_one_has_this_package(ID, Robots0, Robots) :-
    map_values(you_have_not_got_this_package(ID), Robots0, Robots).

:- pred you_have_not_got_this_package(int, int, robot, robot).
:- mode you_have_not_got_this_package(in, in, in, out) is det.

you_have_not_got_this_package(ID, _, Robot0, Robot) :-
    delete(Robot0 ^ carrying, ID, Carrying),
    Robot = Robot0 ^ carrying := Carrying.

%--------------------------------------------------------------------%

:- func should_drop(world) = list(int).

should_drop(World) = Droppings :-
    X = World ^ our_x,
    Y = World ^ our_y,
    IDs = to_sorted_list(World ^ our_carrying),
    Droppings = filter(package_destination(World, X, Y), IDs).

:- pred package_destination(world, int, int, int).
:- mode package_destination(in, in, in, in) is semidet.

package_destination(World, X, Y, ID) :-
    search(World ^ packages, ID, pinfo(X, Y, _)).

%--------------------------------------------------------------------%

:- pred drop(Out, list(int), io, io) <= output(Out).
:- mode drop(in, in, di, uo) is det.

drop(Out, Ids) -->
    stderr_stream(Err),
%    io__write_string(Err, "Dropped\t"),
    write_string(Out, "1 Drop "),
    write_string(Out, join_list(" ", map(int_to_string, Ids))),
    write_string(Out, "\n").

%--------------------------------------------------------------------%

:- pred pick(Out, list(int), io, io) <= output(Out).
:- mode pick(in, in, di, uo) is det.

pick(Out, Ids) -->
    stderr_stream(Err),
%    io__write_string(Err, "Picked\t"),
    write_string(Out, "1 Pick "),
    write_string(Out, join_list(" ", map(int_to_string, Ids))),
    write_string(Out, "\n").

%--------------------------------------------------------------------%

:- pred noop(Out, io, io) <= output(Out).
:- mode noop(in, di, uo) is det.

noop(Out) -->
    stderr_stream(Err),
%    io__write_string(Err, "No-op\t"),
    write_string(Out, "1 Pick 1\n").

%--------------------------------------------------------------------%

:- pred move(Out, dir, io, io) <= output(Out).
:- mode move(in, in, di, uo) is det.

move(Out, Dir) -->
    stderr_stream(Err),
    (
	{ Dir = n },
	write_string(Out, "1 Move N\n")
%	io__write_string(Err, "North\t")
    ;
	{ Dir = s },
	write_string(Out, "1 Move S\n")
%	io__write_string(Err, "South\t")
    ;
	{ Dir = e },
	write_string(Out, "1 Move E\n")
%	io__write_string(Err, "East\t")
    ;
	{ Dir = w },
	write_string(Out, "1 Move W\n")
%	io__write_string(Err, "West\t")
    ).

%--------------------------------------------------------------------%

:- pred time(int, io, io).
:- mode time(out, di, uo) is det.

:- pragma foreign_proc(c, time(T::out, IO0::di, IO::uo),
	[will_not_call_mercury, thread_safe, promise_pure], "
    T = time(NULL);
    IO = IO0;
").

%--------------------------------------------------------------------%

:- end_module world.

